#!/usr/bin/env python3
"""Artistools - light curve functions."""

from .plotlightcurve import main, addargs
from .plotlightcurve import main as plot
from .lightcurve import *
from .viewingangleanalysis import *
